#!/usr/bin/env python
# coding: utf-8


"""

======PROJECT======

MathEnergy

======DESCRIPTION======

The functions in this script are used to modify elements in a Simbench network object.

======Overview functions=====

01: def get_simbench_timestamps(freq = 'qh')

02: def add_timesteps_to_profiles(net, timestamp, steps, 
                                  profile = 'powerplants')

03: def hour_of_year(date, freq = 'h')

04: def update_net(net, p_mw_expan, profiles_abs_new)


​
======AUTHOR, CONTACT & DATE======

Department of Energy and Resource Management, Technical University of Berlin

verwiebe@tu-berlin.de

June 2021

"""



# ---IMPORT STATEMENTS---
import copy

import numpy as np
import pandas as pd

import logging
logger = logging.getLogger('mathenergy')



# ---FUNCTIONS---
# 01
def get_simbench_timestamps(freq = 'qh'):
    """Generates a pandas timestamp index for the simbench dataset.

    Parameters
    ----------
    freq : (str), optional
        Frequency. The default is 'qh'.

    Raises
    ------
    ValueError

    Returns
    -------
    timestamp : (str)
        Timestamp after which steps are added.

    """
    if freq == 'qh':
        timestamp = pd.date_range(start = '1/1/2016', periods = 35136, freq = '0.25H')
    elif freq == 'h':
        timestamp = pd.date_range(start = '1/1/2016', periods = 8784, freq = 'H')
    else:
        raise ValueError

    return timestamp


# 02
def add_timesteps_to_profiles(net, timestamp, steps, profile = 'powerplants'):
    """Correct relative profiles with missing timestamps using linear 
    interpolation.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
    
    timestamp : (str)
        Timestamp after which steps are added.
    
    steps : (int)
        Number of timesteps added.
    
    profile : (str)
        Name of profile to modify ('load', 'powerplants' or 'storages').

    Returns
    -------
    net : (:obj:pandapowerNet)
        Updated pandapower grid object.

    Notes
    -----
    - Only returns profile sums when plot is TRUE.
    - Initial profiles are derived from (:obj:net)
    
    """
    ts_index = [index for index, timestamps in enumerate(net.profiles[profile].time) if timestamp in timestamps][0]
    # Copy first part of dataframe
    df_new_pp_profiles = net.profiles[profile].iloc[:ts_index].copy()
    # Copy remaining dataframe
    df_temp = net.profiles[profile].iloc[ts_index:].copy()
    # Shift 4 rows
    df_temp = df_temp.append([np.nan for x in range(steps)]).shift(steps)
    # Drop line
    df_temp = df_temp.drop(columns = 0)
    # Append
    df_new_pp_profiles = df_new_pp_profiles.append(df_temp)
    # Reset index
    df_new_pp_profiles = df_new_pp_profiles.reset_index(drop = True)
    # Interpolate missing value
    df_new_pp_profiles = df_new_pp_profiles.interpolate(method ='linear',
                                                        limit_direction = 'forward',
                                                        axis = 0)
    # Change profile
    net.profiles[profile] = df_new_pp_profiles
    
    return net


# 03
def hour_of_year(date, freq = 'h'):
    """Compute hour of the year (int) by date.

    Parameters
    ----------
    date : (str)
        Date of the end of the time period.
    
    freq : (str), optional
        Frequency. The default is 'h'.

    Returns
    -------
    hour : (int)
        Hour of the year.

    """
    hour = len(pd.date_range(start = '1/1/2016', end = date, freq = freq))
    
    return hour


# 04
def update_net(net, p_mw_expan, profiles_abs_new):
    """Adjusts parameters of elements in net object. Other functions usually address active power parameters.
	This functions adjusts reactive power and other parameters of elements according to their active power.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
    
    p_mw_expan : (pd.DataFrame)
        Consists power plants that have been expanded or added to the net object but have not been fully parametrized.
    
    profiles_abs_new : (dict)
        Timeseries of generation of active power according to adjustments made by other functions, e.g. distribute_energy_pool().

    Returns
    -------
    None.

    """
    ###############
	# GAS
	###############
	
    # Create dataframe with old and new parameters on the basis of existing dataframe p_mw_expan
    # New_sn_mva is calculated based on initial ratio between p_mw and sn_mva
    # Analog for min_q_mvar and max_q_mvar
    p_mw_expan_all_param = p_mw_expan.assign(sn_mva = net.gen[['sn_mva']],
                                             new_sn_mva = lambda x: ((x['sn_mva']/x['max_p_mw'])*x['new_max_p_mw']), # Scale sn_mva according to increase in new_max_p_mw
                                             min_q_mvar = net.gen[['min_q_mvar']],
                                             new_min_q_mvar = lambda x: ((x['min_q_mvar']/x['max_p_mw'])*x['new_max_p_mw']), # Scale sn_mva according to increase in new_max_p_mw
                                             max_q_mvar = net.gen[['max_q_mvar']],
                                             new_max_q_mvar = lambda x: ((x['max_q_mvar']/x['max_p_mw'])*x['new_max_p_mw']) # Scale sn_mva according to increase in new_max_p_mw                               
                                             )
											 
	# Update net object 
    net.gen.p_mw.update(p_mw_expan_all_param['new_max_p_mw'])
    net.gen.max_p_mw.update(p_mw_expan_all_param['new_max_p_mw']) # Set p_mw and max_p_mw to the same level
    net.gen.sn_mva.update(p_mw_expan_all_param['new_sn_mva'])
    net.gen.min_q_mvar.update(p_mw_expan_all_param['new_min_q_mvar'])
    net.gen.max_q_mvar.update(p_mw_expan_all_param['new_max_q_mvar'])
	
    
	###############
	# OFFSHORE WIND
	###############
	
	# Set apparent power to same level as active power
    net.sgen.p_mw.update(profiles_abs_new[('sgen', 'p_mw')].loc[:,net.sgen.type == 'wind offshore'].max(axis = 0))
    net.sgen.sn_mva.update(net.sgen['p_mw'])
    net.sgen.loc[net.sgen.type == 'wind offshore','profile'] = net.sgen.loc[net.sgen.type == 'wind offshore', 'profile'].index.astype(str) + '_wind_offshore'
	
    
    ###############
	# IMPORTS
	###############
	
	# Imports are defined just like generators with p_mw != sn_mva and sn_mva = max_q_mvar = -min_q_mvar
    # Create new dataframe for import "expansion"
    import_expan_all_param = net.gen[['p_mw', 'max_p_mw' , 'sn_mva', 'min_q_mvar', 'max_q_mvar']].loc[net.gen.type.isin(['imp0', 'imp1'])].copy(deep = True)
	
	# Create dataframe with new values for import parameters
    # new_p_mw from profiles, which have been changed by function distribute_energy_pool()
    # sn_mva is calculated based on initial ratio between p_mw and sn_mva
    # Analog for min_q_mvar and max_q_mvar
    import_expan_all_param = import_expan_all_param.assign(new_p_mw = profiles_abs_new['gen', 'p_mw'][import_expan_all_param.index].max(axis = 0),
                                                           new_sn_mva = lambda x: ((x['sn_mva']/x['p_mw'])*x['new_p_mw']),
                                                           new_min_q_mvar = lambda x: ((x['min_q_mvar']/x['p_mw'])*x['new_p_mw']),
                                                           new_max_q_mvar = lambda x: ((x['max_q_mvar']/x['p_mw'])*x['new_p_mw'])
                                                           )
														   
	# Update net object
    net.gen.p_mw.update(import_expan_all_param['new_p_mw'])
    net.gen.sn_mva.update(import_expan_all_param['new_sn_mva'])
    net.gen.min_q_mvar.update(import_expan_all_param['new_min_q_mvar'])
    net.gen.max_q_mvar.update(import_expan_all_param['new_max_q_mvar'])
	
    
	###############
	# DC LINES
	###############
    
	# Name profiles in net object and columns in normalized profiles
    net.dcline['profile'] = 'profile_' + net.dcline.index.astype(str)
    net.profiles['dcline'] = pd.DataFrame(data = 0, index = net.profiles['load'].index, 
                                          columns = net.dcline.profile)
    net.dcline.p_mw.update(profiles_abs_new[('dcline', 'p_mw')].max(axis = 0))
	
	
    ###############
	# NEW NORMALIZED PROFILES
	###############
  	
    gas_imp_pp = net.gen.loc[net.gen.type.isin(['gas', 'imp0', 'imp1'])].index
    gas_imp_init_profiles = net.gen.loc[gas_imp_pp].profile
  	
    # Create a dataframe for naming new profiles
    # New name includes old profile name and "_adapted_for_" + assigned powerplant ID
    rename_profiles_df = (net.gen.loc[gas_imp_pp].profile.to_frame().assign(new_profile = lambda x: ((x.profile.astype(str)) + '_adapted_for_' + x.index.astype(str))))
							  
	# Update net object with new profile names for generators 
    net.gen.profile.update(rename_profiles_df['new_profile'])
	
	# Add new profiles to net object
    # Normalize them and rename columns at the same time
    net.profiles['powerplants'] = (net.profiles['powerplants'].join(profiles_abs_new[('gen', 'p_mw')].loc[:,gas_imp_pp]
                                                                    .apply(lambda x: x/x.max(), axis = 0)
                                                                    .rename(columns = rename_profiles_df.new_profile)).fillna(0))
								  
    net.profiles['renewables'] = net.profiles['renewables'].join(profiles_abs_new[('sgen', 'p_mw')].loc[:,net.sgen.loc[net.sgen.type == 'wind offshore'].index]
                                                                 .apply(lambda x: x/x.max(), axis = 0)
                                                                 .rename(columns = lambda x: str(x) + '_wind_offshore'))
								  
	# Add new profiles to net object
    # Normalize them and rename columns at the same time
    net.profiles['dcline'] = profiles_abs_new['dcline', 'p_mw'].apply(lambda x: x/x.max(), axis = 0).rename(columns = net.dcline.profile).fillna(0)